Self-supervised learning for Anomaly Detection using 1D-CNN
===========================================================

Identity card
-------------

The identity card is accessible [here](doc/identity_card.yml).

Description
-----------

This package named `cnndrad` is a Python library for detecting anomalies in regularly sampled univariate time series.

The name of the library stands for `1D-CNN Data Reconstruction for Anomaly Detection`.

In short, anomaly detection is based on a two-step method using deep 1D-CNN architectures:
1. Representation learning step: use pretext tasks to learn a representation of anomaly-free data samples in a self-supervised way,
2. Anomaly detection step: raise an anomaly each time the anomaly score of the tested data sample is greater than a threshold.

More details about the method are available in the technical documentation accessible [here](doc/technical_documentation.md).

Prerequisites
-------------

This package is aimed at `Data Scientists` (responsible for providing quality datasets) in collaboration with `Machine Learning Engineers`
(to optimize the parameters related to the training of the underlying Deep Learning model) and `Technical Experts` (expert knowledge is valuable
to set the anomaly threshold and validate the detected anomalies).

As context-level pre-requisites:

- To choose the appropriate sliding window size to extract local substructures from the data, the nature of the time series
  and the dominant period of the signal must be known.

- To learn a good representation of normal data in order to limit the number of false-positive anomaly detections,
  an example of time series without anomaly should be available.

As human-level pre-requisites:

- The user should be familiar with anomaly detection techniques for time series, in particular the use of a sliding window on the input data
  to generate windowed data samples on which the anomaly detection algorithm is applied.
  
- Although it is not required (as the user can use default parameters), some familiarity on the training of deep learning models should be desirable,
  especially for the choice of the component training parameters (batch size, number of epochs, train-validation split of the dataset) and for checking
  the training history (evolution of the loss and validation metric during training) of the underlying deep learning model.

Installation
------------

The following steps are required to install the `cnndrad` library and all the required dependencies:

1. Download the library

```
git clone git@git.irt-systemx.fr:confianceai/ec_5/ec5_as3/1d-cnn-anomaly-detection.git
cd 1d-cnn-anomaly-detection
```

2. Install the required dependencies

```
pip install -r requirements.txt
```

3. Install the library

```
pip install .
```

To check the `cnndrad` library is properly installed, it is advised to run the available test case located in the `tests` directory:

```
cd tests ; ./test_cnndrad.py
```

For more details, please have a look [here](tests/README.md).

Usage
-----

Once installed, you can access the `cnndrad` library directly with the Python `import` statement:

```
import cnndrad
```

The overall picture of the procedure of use can be illustrated as follows:

![procedure](doc/procedure.png "Procedure for using the component")

The procedure of use involves three different steps:

- The first step consists in creating the anomaly detector once the size of the sliding window to be applied on the input data is known.
  This initializes the 1D-CNN model that will be used later to learn a representation of normal data. It is recommended to choose the size of the
  sliding window approximately equal to the number of points corresponding to the period of the data.  If this period is unknown, the component 
  provides the `compute_period` method to give an evaluation.
  
- The second step consists in using anomaly-free data samples (called `train data`) and a set of predefined pretext tasks to learn a
  representation of normal data in a self-supervised way. This is achieved by using the `fit` method that outputs the trained 1D-CNN model
  that was initialized in the first step.
  
- The third step is dedicated to detecting anomalies in the provided data samples (called `test data`). The `score_samples` method
  uses the pretrained 1D-CNN model to reconstruct windowed data samples and to output anomaly scores based on the reconstruction scores that measure
  the distance between the reconstructed windowed data samples and the original ones. These anomaly scores can finally be used by the user to report
  an anomaly whenever the score is above an anomaly threshold set to one by default.
 
Quick start
-----------

For a quick start, we recommend the reader to use the library's default settings and therefore we will only talk about
data-dependent parameters that must be provided.

In the following, we assume that the user wants to detect anomalies in a dataset composed of regularly sampled univariate time series.
This dataset is split into two parts `train_data` (used to learn a representation of the data) and `test_data` (used to detect anomalies).

To use the library, the following steps are required:

### 1. Evaluate the size of the sliding window to be applied on data samples

It is recommended to choose the size of the sliding window approximately equal to the number of points corresponding to the data period.

If not known, the library provides a method to evaluate the data period and the associated half width confidence interval:

```
period, confidence = cnndrad.utils.Signal.compute_period(train_data)
```

Anomaly detection is based on the signal obtained by applying the sliding window on the data samples.
By default, the windowed data samples are split into three parts, with each part reconstructed by a deep neural network.
As a result, detection is optimal when the size of the sliding windows is close to the data period while being divisible by 3.

The size of the sliding window is finally computed as follows:

```
import numpy as np
window_size = int(np.ceil(period / 3) * 3)
```

### 2. Create the anomaly detector

Once the size of the sliding windows is chosen, the anomaly detector can be created:

```
anom_detector = cnndrad.DataReconstructionAD(window_size)
```

### 3. Make the detector learn a representation of normal data samples

In this section, the `train_data` is assumed not to contain anomalies.

To detect anomalies, the detector must first learn what normal data samples look like.

This is achieved using the `fit` method:

```
anom_detector.fit(train_data)
```

The detector is now ready to detect anomalous data samples.


### 4. Use the detector to find anomalies in data samples

In this section, the `test_data` is assumed to contain anomalies.

To detect anomalies, we first compute anomaly scores:

```
score = - anom_detector.score_samples(test_data)
```

The `score` contains positive real values.

The higher the score values, the higher the probability that there is an anomaly.

Usually, we consider that there is an anomaly when `score` values are greater than an anomaly threshold, which is by default set to 1.


Finally, anomaly segments can also be computed:

```
anomaly_segments = anom_detector.anomaly_segments(test_data)
```

Example
-------

The `cnndrad` library comes with an example of use located in the `examples` directory.

For more details, please have a look [here](examples/README.md).

Demonstration
-------------

The `cnndrad` library comes with a demonstrator located in the `demonstrators` directory.

For more details, please have a look [here](demonstrators/README.md).

Validation
----------

The `cnndrad` library has been validated against a set of 250 public datasets.

For more details, please have a look [here](validation/README.md).

License
-------

Copyright (c) 2021-2023 by CEA. All rights reserved.
