import math
import numpy as np

from tensorflow import keras


class WindowedData(keras.utils.Sequence):
    
    def __init__(self, X, window_size, batch_size, window_stride=1, training=False, shuffle=False):
        """
        ----------------------
        GENERATE WINDOWED DATA
        ----------------------
        
        :param array, X: Array of shape (sample_size, 2) where columns are (timestamps, values).
        :param int, window_size: Size of the sliding window applied on data samples.
        :param int, batch_size: Number of samples per batch to train/evaluate the model.
        :param int, window_stride:  The stride of the sliding window.
        :param bool, training: Set to True only during training.
        :param bool, shuffle : If True, the data samples will be shuffled before each epoch during training.
        
        """
        self.X = X
        self.window_size = window_size
        self.batch_size = batch_size
        self.training = training
        self.shuffle = shuffle
        
        nb_max_windows = max(len(X) - self.window_size + 1, 0)
        self.indexes = np.arange(0, nb_max_windows, window_stride)
        self.nb_windows = len(self.indexes)
    
    def __len__(self):
        return math.ceil(self.nb_windows / self.batch_size)
    
    def __getitem__(self, idx):
        low = idx * self.batch_size
        high = min(low + self.batch_size, self.nb_windows)
                
        windows_l = []
        for i in range(low, high):
            j = self.indexes[i]
            windows_l.append(self.X[j:j + self.window_size, 1])
        windows = np.array(windows_l).astype('float32')
        
        if self.training:
            return windows, windows
        else:
            return windows
        
    def on_epoch_end(self):
        if self.shuffle:
            np.random.shuffle(self.indexes)
