import os
from pathlib import Path
import json

import numpy as np
from scipy.interpolate import interp1d
import pandas as pd

import matplotlib.pyplot as plt

import plotly
from plotly.subplots import make_subplots
import plotly.graph_objects as go

import tensorflow as tf
from tensorflow import keras
from tensorflow.python.profiler.model_analyzer import profile
from tensorflow.python.profiler.option_builder import ProfileOptionBuilder
from tensorflow.python.framework.convert_to_constants import convert_variables_to_constants_v2_as_graph

from .models import CNN_1D_3x3Conv
from .data import WindowedData
from .utils import Plot


def rmse(y_true, y_pred):
    """Define Root Mean Squared Error metric"""
    return tf.sqrt(keras.losses.mean_squared_error(y_true, y_pred))

keras.utils.get_custom_objects().update({'rmse': rmse})


class DataReconstructionAD(object):
    
    _models = { 'CNN_1D_3x3Conv' : CNN_1D_3x3Conv() }
    
    def __init__(self, window_size, window_stride=1, reconstruct=[True]*3,
                       model_name='CNN_1D_3x3Conv', metric='mae',
                       batch_size=32, epochs=100, validation_split=0.2,
                       work_dir='./', device='/gpu:0', **kwargs):
        """
        -------------------------------------------------------------
        ANOMALY DETECTION USING DATA RECONSTRUCTION AS A PRETEXT TASK
        -------------------------------------------------------------
        
        :param int, window_size: Size of the sliding window applied on data samples.
        :param int, window_stride:  The stride of the sliding window.
        :param [bool], reconstruct: Windowed data samples are split into parts. List of boolean values to specify which part of the windowed data samples will be recontructed.
        :param str, model_name: Name of the model used to reconstruction data samples.
        :param str, metric: Metric used to compare the quality of the resconstructions against the origianl data samples. Can be 'mae' (Mean Absolute Error) or 'rmse' (Root Mean Squared Error).
        :param int, batch_size: Number of samples per batch to train/evaluate the model.
        :param int, epochs: Number of epochs to train the model.
        :param float, validation_split: Value between 0 and 1. Fraction of the data to be used as validation data.
        :param str, work_dir: Working directory used to save model parameters and figures.
        :param str, device: Device used to train/evaluate the model.
        
        To learn a data representation
        >>> anom_detector.fit(train_data)
        >>> anom_detector.plot_history()
        >>> anom_detector.plot_metric_windows_distribution(train_data)
        
        To compare model performance with baseline
        >>> anom_detector.compare_model_performance_with_baseline(train_data)
        
        To compute anomaly score
        >>> anomaly_score = anom_detector.score_samples(test_data)
        >>> anom_detector.plot_score_samples(test_data, savefig=True)
        >>> anom_detector.plot_score_windows_distribution(test_data)
        >>> anom_detector.plot_data_windows_reconstruction(test_data)
        
        To compute anomaly segments
        >>> anomaly_segments = anom_detector.anomaly_segments(test_data)
        
        """
        super(DataReconstructionAD, self).__init__()
        
        self.window_size = window_size
        self.window_stride = window_stride
        self.reconstruct = reconstruct
        
        assert model_name in self._models
        self.model_name = model_name
        assert metric in ['mae', 'rmse']
        self.metric = metric
        
        self.batch_size = batch_size
        self.epochs = epochs
        self.validation_split = validation_split
        
        self.work_dir = Path(work_dir)
        self.model_dir = self.work_dir / 'model'
        self._filename_weights = self.model_dir / 'weights.h5'
        self._filename_history = self.model_dir / 'history.json'
        self._filename_parameters = self.model_dir / 'parameters.json'
        self.output_dir = self.work_dir / 'output'
        
        self.device = device
        self._kwargs = kwargs
        
        # Initilaize
        
        self._model = None
        self._history = None
        self._params = { 'mean' : None, 'std' : None, 'threshold' : None }
    
    def __repr__(self):
        return f"{self.__class__.__name__}(window_size={self.window_size}, window_stride={self.window_stride}, reconstruct={self.reconstruct})"
    
# =======================================================================================================================
# MODEL ASSOCIATED WITH THE PRETEXT TASK
# =======================================================================================================================
    
    def _build_model(self):
        """Create model"""
        model = self._models[self.model_name].build_model(input_size=self.window_size, reconstruct=self.reconstruct, **self._kwargs)
        model.compile(optimizer='adam', loss='mse', metrics=[self.metric])
        return model
    
    def load_model(self):
        """Load model, model history and related parameters"""
        # Load model
        if os.path.isfile(self._filename_weights):
            with tf.device(self.device):
                
                # Create model
                self._model = self._build_model()
                
                # Load best model
                self._model.load_weights(self._filename_weights)
        
        # Load history
        if os.path.isfile(self._filename_history):
            with open(self._filename_history, 'r') as f:
                self._history = json.load(f)
        
        # Load parameters
        if os.path.isfile(self._filename_parameters):
            with open(self._filename_parameters, 'r') as f:
                self._params = json.load(f)
    
    def save_model(self):
        """Save model parameters"""
        # Save parameters
        with open(self._filename_parameters, 'w') as f:
            json.dump(self._params, f)
            
    def _model_FLOPs(self, batch_size=1):
        """Compute model FLOPs"""
        with tf.device(self.device):
            # Convert model into frozen graph
            inputs = [tf.TensorSpec(shape=(batch_size,) + input.shape[1:]) for input in self._model.inputs]
            real_model = tf.function(self._model).get_concrete_function(inputs)
            frozen_func, _ = convert_variables_to_constants_v2_as_graph(real_model)
            
            # Profile frozen graph
            opts = ProfileOptionBuilder(ProfileOptionBuilder.float_operation()).with_empty_output().build()
            graph_info = profile(frozen_func.graph, options=opts)
            
            # The //2 is necessary since `profile` counts multiply and accumulate
            # as two flops, here we report the total number of multiply accumulate ops
            flops = graph_info.total_float_ops // 2
        
        return flops
    
    def print_model(self):
        """Print model summary and FLOPs"""
        if self._model is not None:
            self._model.summary()
            print(f'Model FLOPs: {self._model_FLOPs():,}')
        else:
            print('No model available.')
    
# =======================================================================================================================
# LEARN A DATA REPRESENTATION
# =======================================================================================================================
    
# -----------------------------------------------------------------------------------------------------------------------
    def fit(self, X, y=None):
        """
        Fit the model to learn a data representation
        
        :param array, X: Ndarray or Pandas DataFrame of shape (sample_size, 2) where columns are (timestamps, values).
        
        """
        # Prepare data
        X = self._prepare_data(X, update_params=True)
        
        # Train model
        self._train_model(X)
        
        # Compute threshold
        self._compute_threshold(X)
        
        # Save model parameters
        self.save_model()
        
        return self
    
# -----------------------------------------------------------------------------------------------------------------------
    
    def _validate_data(self, X, epsilon=0.01):
        """Check and convert data format to ndarray"""
        # Convert input data to ndarray (if needed)
        if isinstance(X, pd.DataFrame):
            if isinstance(X.index, pd.DatetimeIndex):
                X = X.reset_index()
            X = X.to_numpy()
        elif isinstance(X, np.ndarray):
            X = np.copy(X)
        else:
            raise Exception('Wrong input data type')
        
        # Check ndarray dimensionality
        if not (X.ndim == 2 and X.shape[1] == 2):
            raise Exception('Wrong input data format')
        
        # Check that input data is regularly sampled
        delta = np.diff(X[:,0])
        delta /= np.mean(delta)
        if np.any((delta < 1 - epsilon) | (delta > 1 + epsilon)):
            raise Exception('Input data must be regularly sampled')
        
        return X
    
    def _prepare_data(self, X, update_params=False):
        """Validate and standardize data"""
        # Validate data
        X = self._validate_data(X)
        
        # Update data mean and std
        if update_params:
            self._params['mean'], self._params['std'] = np.mean(X[:,1]), np.std(X[:,1])
        
        # Standardize data
        X[:,1] -= self._params['mean']
        X[:,1] /= self._params['std']
        
        return X
    
    def _train_model(self, X):
        """Train model, save history and compute metrics after training"""
        # Create training and validation sets
        index = int(self.validation_split * len(X))
        index = max(index, self.window_size)
        if len(X) < index + self.window_size:
            raise Exception('Not enough input data samples')
        X_train, X_valid = X[:-index], X[-index:]
        
        X_train_windows = WindowedData(X_train, window_size=self.window_size, batch_size=self.batch_size, window_stride=self.window_stride, training=True, shuffle=True)
        X_valid_windows = WindowedData(X_valid, window_size=self.window_size, batch_size=self.batch_size, window_stride=self.window_stride, training=True, shuffle=False)
        
        # Train model
        with tf.device(self.device):
            
            # Create model
            self._model = self._build_model()
            
            # Fit model
            os.makedirs(os.path.dirname(self._filename_weights), exist_ok=True)
            callbacks = [ keras.callbacks.ModelCheckpoint(filepath=self._filename_weights,
                                                          monitor=f'val_{self.metric}', mode='min',
                                                          save_weights_only=True, save_best_only=True,
                                                          verbose=0) ]
            
            history = self._model.fit(x=X_train_windows,
                                     epochs=self.epochs,
                                     validation_data=X_valid_windows,
                                     callbacks=callbacks)
            
            # Save history            
            self._history = history.history
            os.makedirs(os.path.dirname(self._filename_history), exist_ok=True)
            with open(self._filename_history, 'w') as f:
                json.dump(self._history, f)
            
            # Load best model
            self._model.load_weights(self._filename_weights)
            
            # Compute best model metrics
            loss, metric = self._model.evaluate(X_train_windows, verbose=0)
            val_loss, val_metric = self._model.evaluate(X_valid_windows, verbose=0)
            print(f'loss: {loss:.4f} - {self.metric}: {metric:.4f}')
            print(f'val_loss: {val_loss:.4f} - val_{self.metric}: {val_metric:.4f}')
    
    def plot_history(self, smooth_factor=0., savefig=False):
        """Plot model history during training"""
        if not self._history:
            return
        
        # Loss
        smooth_loss = Plot.smooth_curve(self._history['loss'], factor=smooth_factor)
        smooth_val_loss = Plot.smooth_curve(self._history['val_loss'], factor=smooth_factor)
        plt.plot(smooth_loss, label='Training Loss')
        plt.plot(smooth_val_loss, label='Validation Loss')
        plt.xlabel('Epochs')
        plt.ylabel('Loss')
        plt.yscale('log')
        plt.legend()
        
        if savefig:
            filename = self.output_dir / 'history_loss.png'
            os.makedirs(os.path.dirname(filename), exist_ok=True)
            plt.savefig(filename, bbox_inches='tight')
        else:
            plt.show()
        plt.close()
        
        # Metric
        smooth_metric = Plot.smooth_curve(self._history[self.metric], factor=smooth_factor)
        smooth_val_metric = Plot.smooth_curve(self._history[f'val_{self.metric}'], factor=smooth_factor)
        plt.plot(smooth_metric, label=f'Training {self.metric}')
        plt.plot(smooth_val_metric, label=f'Validation {self.metric}')
        plt.xlabel('Epochs')
        plt.ylabel(self.metric)
        plt.yscale('log')
        plt.legend()
        
        if savefig:
            filename = self.output_dir / 'history_metric.png'
            os.makedirs(os.path.dirname(filename), exist_ok=True)
            plt.savefig(filename, bbox_inches='tight')
        else:
            plt.show()
        plt.close()
    
    def _baseline(self, windows, kind='const'):
        """
        Compute reconstruction baseline to compare with trained model
        
        :param list, windows: List of windowed data.
        :param str, kind: Type of reconstruction. It can be 'const', 'linear', 'cubic' or 'quadratic' interpolation.
        :return list: List of reconstructed windowed data using baseline.
        
        """
        reconstructions = []
        
        assert kind in ['const', 'linear', 'cubic', 'quadratic']
        
        if kind == 'const':
            
            window_split = len(self.reconstruct)
            sample_size = self.window_size // window_split
            
            for window in windows:
                reconstruction = window.copy()
                mean = np.mean(np.array_split(window[:window_split*sample_size], window_split), axis=-1)
                
                for i, reconstruct in enumerate(self.reconstruct):
                    if reconstruct:
                        reconstruction[i*sample_size:(i+1)*sample_size] = mean[i] - mean.sum()
                
                reconstructions.append(reconstruction)
        
        elif kind in ['linear', 'cubic', 'quadratic']:
            
            window_split = len(self.reconstruct)
            sample_size = self.window_size // window_split
            
            for window in windows:
                reconstruction = window.copy()
                
                for i, reconstruct in enumerate(self.reconstruct):
                    if reconstruct:
                        if i == 0:
                            reconstruction[i*sample_size:(i+1)*sample_size] = window[(i+1)*sample_size]
                        elif i == window_split-1:
                            reconstruction[i*sample_size:(i+1)*sample_size] = window[i*sample_size-1]
                        else:
                            x = np.concatenate([np.arange((i-1)*sample_size, i*sample_size), np.arange((i+1)*sample_size, (i+2)*sample_size)])
                            f_interp = interp1d(x, window[x], kind=kind)
                            
                            x_interp = np.arange(i*sample_size, (i+1)*sample_size)
                            reconstruction[i*sample_size:(i+1)*sample_size] = f_interp(x_interp)
                
                reconstructions.append(reconstruction)
        
        return reconstructions
    
    def _reconstruct(self, windows, baseline=False):
        """
        Reconstruct windowed data
        
        :param list, windows: List of windowed data.
        :param bool, baseline: If True, the baseline is computed instead of the model prediction.
        :return list: List of reconstructed windowed data.
        
        """
        if baseline:
            reconstructions = self._baseline(windows)
        else:
            with tf.device(self.device):
                reconstructions = self._model.predict_on_batch(windows)
        return reconstructions
    
    def _compute_metric_windows(self, X_windows, baseline=False):
        """Compute metric for to the reconstructed windowed data"""
        
        with tf.device(self.device):
            metric_function = keras.metrics.get(self.metric)
            
            metric_windows = []
            for windows in X_windows:
                reconstructions = self._reconstruct(windows, baseline=baseline)
                metric = metric_function(reconstructions, windows).numpy()
                metric_windows.append(metric)
            metric_windows = np.concatenate(metric_windows)
        
        return metric_windows
    
    def plot_metric_windows_distribution(self, X, baseline=False, savefig=False):
        """
        Plot metric distribution of the reconstructed windowed data
        
        :param array, X: Ndarray or Pandas DataFrame of shape (sample_size, 2) where columns are (timestamps, values).
        :param bool, baseline: If True, use baseline instead of the model to reconstruct windowed data.
        :param bool, savefig: If True, save figure as PNG file.
        
        """
        X = self._prepare_data(X)
        X_windows = WindowedData(X, window_size=self.window_size, batch_size=self.batch_size, training=False, shuffle=False)
        metric_windows = self._compute_metric_windows(X_windows, baseline=baseline)
        
        plt.hist(metric_windows, bins=50)
        if self._params['threshold']:
            plt.axvline(self._params['threshold'], 0, 1, color='g', linestyle='--', label='Threshold')
        plt.xlabel(self.metric)
        plt.ylabel('No of examples')
        if savefig:
            filename = self.output_dir / 'metric_windows_distribution.png'
            os.makedirs(os.path.dirname(filename), exist_ok=True)
            plt.savefig(filename, bbox_inches='tight')
        else:
            plt.show()
        plt.close()
    
    def _compute_threshold(self, X, q=99):
        """
        Compute anomaly threshold relative to the data.
        For a metric value above this threshold, a reconstrution error is considered an anomaly.
        
        :param array, X: Array of shape (sample_size, 2) where columns are (timestamps, values).
        :param float, q: Percentile of the metric values corresponding to the anomaly threshold.
        
        """
        X_windows = WindowedData(X, window_size=self.window_size, batch_size=self.batch_size, training=False, shuffle=False)
        metric_windows = self._compute_metric_windows(X_windows, baseline=False)
        self._params['threshold'] = np.percentile(metric_windows, q)
        print(f"threshold: {self._params['threshold']:.4f}")
    
    def compare_model_performance_with_baseline(self, X):
        """
        Compare model performance with baseline
        
        :param array, X: Ndarray or Pandas DataFrame of shape (sample_size, 2) where columns are (timestamps, values).
        
        """
        X = self._prepare_data(X)
        X_windows = WindowedData(X, window_size=self.window_size, batch_size=self.batch_size, training=False, shuffle=False)
        
        print('Compare model performance with baseline')
        
        if self._model:
            metric_model = np.mean(self._compute_metric_windows(X_windows, baseline=False))
            print(f'\tmodel - {self.metric}: {metric_model:.4f}')
        else:
            print(f'\tmodel - No model available.')
        
        metric_baseline = np.mean(self._compute_metric_windows(X_windows, baseline=True))
        print(f'\tbaseline - {self.metric}: {metric_baseline:.4f}')
    
# =======================================================================================================================
# ANOMALY DETECTION
# =======================================================================================================================
    
    def _compute_score_windows(self, X_windows, baseline=False):
        """Compute anomaly score of the reconstructed windowed data"""
        metric_windows = self._compute_metric_windows(X_windows, baseline=baseline)
        return metric_windows / self._params['threshold']
    
    def plot_score_windows_distribution(self, X, baseline=False, savefig=False, savename='score_windows_distribution'):
        """
        Plot anomaly score distribution of the reconstructed windowed data
        
        :param array, X: Ndarray or Pandas DataFrame of shape (sample_size, 2) where columns are (timestamps, values).
        :param bool, baseline: If True, use baseline instead of the model to reconstruct windowed data.
        :param bool, savefig: If True, save figure as PNG file.
        :param str, savename: Name of the saved figure.
        
        """
        X = self._prepare_data(X)
        X_windows = WindowedData(X, window_size=self.window_size, batch_size=self.batch_size, training=False, shuffle=False)
        score_windows = self._compute_score_windows(X_windows, baseline=baseline)
        
        plt.hist(score_windows, bins=50)
        plt.axvline(1., 0, 1, color='g', linestyle='--', label='Threshold')
        plt.xlabel('Score')
        plt.ylabel('No of examples')
        if savefig:
            filename = self.output_dir / f'{savename}.png'
            os.makedirs(os.path.dirname(filename), exist_ok=True)
            plt.savefig(filename, bbox_inches='tight')
        else:
            plt.show()
        plt.close()
    
    def plot_data_windows_reconstruction(self, X, baseline=False, min_score=None, sort=False, reverse=True):
        """
        Plot the reconstructed windowed data
        
        :param array, X: Ndarray or Pandas DataFrame of shape (sample_size, 2) where columns are (timestamps, values).
        :param bool, baseline: If True, use baseline instead of the model to reconstruct windowed data.
        :param float, min_score: Minimum anomaly score value for which the reconstructed windowed data is plot.
        :param bool, sort: If True, sort the reconstructed data w.r.t. the metric values.
        :param bool, reverse: If True (resp. False), sort metric values in descending (resp. ascending) order.
        
        """
        X = self._prepare_data(X)
        X_windows1 = WindowedData(X, window_size=self.window_size, batch_size=self.batch_size, training=False, shuffle=False)
        score_windows = self._compute_score_windows(X_windows1, baseline=baseline)
        
        # Window selection        
        if min_score is None:
            selection = np.arange(len(score_windows))
            stride = self.window_size
        else:
            selection = np.where(score_windows > min_score)[0]
            if sort:
                order = np.argsort(score_windows[selection])
                if reverse:
                    order = order[::-1]
                selection = selection[order]
            stride = 1
        
        # Window data
        X_windows2 = WindowedData(X, window_size=self.window_size, batch_size=1, training=False, shuffle=False)
        index_max = len(selection)
        
        def get_window_data(index):
            k = selection[index]
            windows = X_windows2[k]
            reconstructions = self._reconstruct(windows, baseline=baseline)
            return self._params['mean'] + self._params['std'] * windows[0], \
                   self._params['mean'] + self._params['std'] * reconstructions[0], \
                   score_windows[k]
        
        # Plot recontructions
        def plot_window_data(ymin, ymax, stride=1, step=5):
            def handle(event):
                nonlocal index, stride
                if event.key == 'right':
                    index = (index + stride) % index_max
                    draw()
                elif event.key == 'left':
                    index = (index - stride) % index_max
                    draw()
                elif event.key == 'up':
                    stride = min(stride + step, self.window_size)
                    draw()
                elif event.key == 'down':
                    stride = max(stride - step, 1)
                    draw()
                elif event.key == 'escape':
                    plt.close()
            
            fig = plt.figure()
            fig.canvas.mpl_connect('key_press_event', handle)
            
            index = 0
            l1, = plt.plot(np.arange(self.window_size), np.zeros(self.window_size), 'g')
            l2, = plt.plot(np.arange(self.window_size), np.zeros(self.window_size), 'r')
            
            def draw():            
                ground_truth, reconstruction, score = get_window_data(index)
                l1.set_ydata(ground_truth)
                l2.set_ydata(reconstruction)
                plt.suptitle('Data reconstruction')
                plt.title(f'{index + 1}/{index_max} (+{stride}) - Score: {score:.4f}')
                plt.legend(handles=[l1, l2],labels=['Ground Truth', 'Reconstruction'])
                plt.ylim(ymin, ymax)
                plt.draw()
            
            draw()
            plt.show()
        
        X_min = self._params['mean'] + self._params['std'] * np.floor(np.min(X[:,1]))
        X_max = self._params['mean'] + self._params['std'] * np.ceil(np.max(X[:,1]))
        
        plot_window_data(X_min, X_max, stride=stride)
    
    def _compute_score_samples(self, X, average=True):
        """
        Compute anomaly score for data samples
        
        :param array, X: Array of shape (sample_size, 2) where columns are (timestamps, values).
        :param bool, Average: if True, average anomaly score to take into account overlapping sliding windows.
        :return array: Array of shape (sample_size).
        
        """
        X_windows = WindowedData(X, window_size=self.window_size, batch_size=self.batch_size, training=False, shuffle=False)
        score_windows = self._compute_score_windows(X_windows)
        
        # Locate the values at center of the windows
        after = (self.window_size - 1) // 2
        before = (self.window_size - 1) - after
        score = np.pad(score_windows, (before, after), mode='edge')
        
        # Average score to take into account overlapping windows
        if average:
            score = pd.DataFrame(score).rolling(self.window_size, center=True, min_periods=1).mean().iloc[:,-1].to_numpy()
        
        return score
    
# -----------------------------------------------------------------------------------------------------------------------
    def score_samples(self, X):
        """
        Compute anomaly score for data samples
        
        :param array, X: Ndarray or Pandas DataFrame of shape (sample_size, 2) where columns are (timestamps, values).
        :return array: Array of shape (sample_size).
        
        """
        X = self._prepare_data(X)
        score = self._compute_score_samples(X)
        return -score
# -----------------------------------------------------------------------------------------------------------------------        
    
    def plot_score_samples(self, X, savefig=False, savename='score_samples'):
        """
        Plot anomaly score for data samples
        
        :param array, X: Array of shape (sample_size, 2) where columns are (timestamps, values).
        :param bool, savefig: If True, save figure as HTML file.
        :param str, savename: Name of the saved figure.
        
        """
        X = self._prepare_data(X)
        score = self._compute_score_samples(X)
        
        fig = make_subplots(rows=2, cols=1)
        y = self._params['mean'] + self._params['std'] * X[:,1]
        fig.append_trace(go.Scatter(x=X[:,0], y=y, mode='lines', name='Dataset'), row=1, col=1)
        fig.append_trace(go.Scatter(x=X[:,0], y=score, mode='lines', name='Anomaly score'), row=2, col=1)
        fig.update_xaxes(fixedrange=False, rangeslider_visible=False, row=1, col=1)
        fig.update_xaxes(fixedrange=False, rangeslider_visible=True, row=2, col=1)
        fig.update_xaxes(matches='x')
        fig.update_layout(title=f'Sliding window size: {self.window_size}')
        
        filename = self.output_dir / f'{savename}.html'
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        plotly.offline.plot(fig, filename=str(filename), include_plotlyjs='cdn', auto_open=False)
    
    def anomaly_segments(self, X, min_score=1., savefile=False, savename='anomaly_segments'):
        """
        Compute anomaly segments for data samples
        
        :param array, X: Array of shape (sample_size, 2) where columns are (timestamps, values).
        :param float, min_score: Minimum anomaly score value for which the anomaly segments are computed.
        :param bool, savefile: If True, save anomaly segments as CSV file.
        :param bool, savename: Name of the CSV file.
        :return Array: Array of shape (segments_size, 3) where columns are (anomaly start time, anomaly end time, anomaly duration).
        
        """
        X = self._prepare_data(X)
        score = self._compute_score_samples(X)
        score = np.column_stack([X[:,0], score])
        
        df = pd.DataFrame(score, columns=['timestamp','score'])
        
        is_anomaly = df['score'].gt(min_score)
        groups = (~is_anomaly).cumsum()
        
        anomalies = df.groupby(groups).agg(anomaly_start=('timestamp', 'min'), anomaly_end=('timestamp', 'max'), count=('score', 'count')).rename_axis(None, axis=0)
        anomalies = anomalies[anomalies['count'] > 2].drop(['count'], axis=1).reset_index(drop=True)
        anomalies['duration'] = anomalies['anomaly_end'] - anomalies['anomaly_start']
        
        columns = ['anomaly_start', 'anomaly_end', 'duration']
        anomalies = anomalies[columns]
        
        if savefile:
            filename = self.output_dir / f'{savename}.csv'
            os.makedirs(os.path.dirname(filename), exist_ok=True)
            anomalies.to_csv(filename, index=False, sep=';')
        
        return anomalies.to_numpy()
