import numpy as np

import tensorflow as tf
from tensorflow import keras


class CNN_1D_3x3Conv(object):
    
    @classmethod
    def build_model(cls, input_size, reconstruct=[True]*3, nb_filters=[32, 32, 32], activation='relu', max_pooling=False, **kwargs):
    
        """
        -----------------------------------------------
        BUILD 1D-CNN MODEL USED FOR DATA RECONSTRUCTION
        -----------------------------------------------
        
        :param int, input_size: Size of the input data.
        :param [bool], reconstruct: Windowed input data is split into parts. List of boolean values to decide which part of the windowed data will be recontructed.
        :param [int], nb_filters: The model uses 3 convolutional blocks. List of number of filters to use in each convolutional block.
        :param str, activation: The activation function to use in the model.
        :param bool, max_pooling: if True (resp. False), MaxPooling (resp. strided convolution) is used in the convolutional blocks.
        
        """
        # Check input parameters
        
        split = len(reconstruct)
        sample_size = input_size // split
        assert isinstance(nb_filters, list) and len(nb_filters) == 3
        
        # Define encoder
        
        sample = keras.layers.Input(shape=(sample_size,))
        x = keras.layers.Reshape((sample_size, 1)) (sample)
        output_padding_l = []
        for filters in nb_filters:
            x = keras.layers.Conv1D(filters, kernel_size=3, padding='same', activation=activation) (x)
            x = keras.layers.Conv1D(filters, kernel_size=3, padding='same', activation=activation) (x)
            output_padding = 1 if (x.shape[1] % 2) == 0 else 0
            output_padding_l.append(output_padding)
            if max_pooling:
                x = keras.layers.Conv1D(filters, kernel_size=3, padding='same', activation=activation) (x)
                x = keras.layers.MaxPooling1D(pool_size=2, padding='same') (x)
            else:
                x = keras.layers.Conv1D(filters, kernel_size=3, padding='same', activation=activation, strides=2) (x)
        feature_shape = x.shape[1:]
        feature_size = np.prod(feature_shape)
        encoded = keras.layers.Flatten() (x)
        
        encoder = keras.Model(inputs=[sample], outputs=[encoded], name='encoder') 
        
        # Define decoder
        
        feature = keras.layers.Input(shape=(feature_size,))
        y = keras.layers.Reshape(feature_shape) (feature)
        for filters, output_padding in zip(nb_filters[::-1], output_padding_l[::-1]):
            y = keras.layers.Conv1DTranspose(filters, kernel_size=3, padding='same', activation=activation, strides=2, output_padding=output_padding) (y)
        y = keras.layers.Conv1D(1, kernel_size=3, padding='same') (y)
        decoded = keras.layers.Reshape((sample_size,)) (y)
        
        decoder = keras.Model(inputs=[feature], outputs=[decoded], name='decoder')
        
        # Define model
        
        input = keras.layers.Input(shape=(input_size,))
        size_splits = [sample_size] * split + [input_size % split]
        samples = keras.layers.Lambda(lambda x: tf.split(x, size_splits, axis=1)) (input)
        features = [encoder(sample) for sample in samples[:split]]
        
        output_l = []
        for i in range(split):
            if reconstruct[i]:
                z = keras.layers.Concatenate() (features[:i] + features[i+1:])
                feature = keras.layers.Dense(feature_size, activation=activation) (z)
                output_l.append(decoder(feature))
            else:
                output_l.append(samples[i])
        output_l.append(samples[split])
        output = keras.layers.Concatenate() (output_l)
        
        name = f"{cls.__name__}_{'_'.join(map(str, nb_filters))}"
        name += "_with_MaxPooling" if max_pooling else "_with_Stride"        
        model = keras.Model(inputs=[input], outputs=[output], name=name)
        return model


if __name__ == '__main__':
    
    input_size = 405
    split = 3
    reconstruct = [True] * split
    max_pooling = False
    
    CNN_1D_3x3Conv().build_model(input_size=input_size, reconstruct=reconstruct, max_pooling=max_pooling).summary()
