import os

import numpy as np
import scipy.signal
from scipy.stats import t
import pandas as pd

from matplotlib import pyplot as plt


class Signal():
    
    @staticmethod
    def compute_period(X, plot=False, savefig=False, filename='period.png'):
        """
        ---------------------
        COMPUTE SIGNAL PERIOD
        ---------------------
        
        :param array, X: Ndarray or Pandas DataFrame of shape (sample_size, 2) where columns are (timestamps, values).
        :param bool, plot: If True, plot autocorrelation and relevant peaks.
        :param bool, savefig: If True, save figure as PNG file.
        :return float, float: Signal period and the associated half width confidence interval.
        
        """
        # Get signal values
        x = Signal._get_signal_values(X)
        
        # Center values
        x -= x.mean()
        
        # Compute the autocorrelation
        xcorr = scipy.signal.correlate(x, x, mode='same')
        xcorr = xcorr[len(xcorr)//2:] / (len(x) * x.var())
        
        # Compute the minimum distance between neighbouring peaks
        found, _ = scipy.signal.find_peaks(np.concatenate(([np.min(xcorr)], xcorr)), height=0.1, prominence=0.025)
        found -= 1
        ind = np.argsort(xcorr[found])[::-1]
        distance = 0.95 * (found[ind[1]] - found[ind[0]])
        
        # Find peak candidates
        found, _ = scipy.signal.find_peaks(np.concatenate(([np.min(xcorr)], xcorr)), height=0.1, prominence=0.025, distance=distance)
        found -= 1
        
        # Select peaks
        peaks = found[:2].tolist()
        period = found[1] - found[0]
        for peak in found[2:]:
            delta = peak - peaks[-1]
            if delta < 0.95 * period:
                # Skip peak candidate
                continue
            elif delta > 1.05 * period:
                # Skip all remaining peak candidates
                break
            peaks.append(peak)
        
        # Plot autocorrelation and peaks
        if plot:
            Signal._plot_autocorrelation(xcorr, peaks, savefig=savefig, filename=filename)
        
        # Compute period
        delta = np.diff(peaks)
        period = np.mean(delta)
        n = len(delta)
        if n > 1:
            std = np.std(delta, ddof=1)
            confidence = Signal._confidence_interval_half_width(n, std)
        else:
            confidence = 0
        return period, confidence
    
    @staticmethod
    def _get_signal_values(X):
        """Return signal values from input data"""
        if isinstance(X, pd.DataFrame):
            if isinstance(X.index, pd.DatetimeIndex):
                X = X.reset_index()
            X = X.to_numpy()
        elif isinstance(X, np.ndarray):
            X = np.copy(X)
        else:
            raise Exception('Wrong input data type')
        
        if not (X.ndim == 2 and X.shape[1] == 2):
            raise Exception('Wrong input data format')
        
        return X[:,1]
    
    @staticmethod
    def _confidence_interval_half_width(n, std, confidence_level=0.95):
        """Compute half width of confidence interval"""
        # See https://en.wikipedia.org/wiki/Confidence_interval
        proba = 1 - (1 - confidence_level) / 2
        return std / np.sqrt(n) * t.ppf(proba, n - 1)
    
    @staticmethod
    def _plot_autocorrelation(xcorr, peaks, savefig=False, filename='period.png'):
        """Plot autocorrelation function with found peaks"""
        fig = plt.figure()
        
        # Plot
        ax = fig.add_subplot()
        ax.plot(xcorr)
        ax.plot(peaks, xcorr[peaks], 'xr')
        
        ax.set_xlabel('Lag', fontsize=20)
        ax.set_ylabel('Autocorrelation', fontsize=20)
        
        fig.tight_layout()
        
        if savefig:
            os.makedirs(os.path.dirname(filename), exist_ok=True)
            plt.savefig(filename)
        else:
            plt.show()
        plt.close()
