#!/usr/bin/env python3

import os, sys
import shutil
from pathlib import Path

import numpy as np

import plotly
from plotly.subplots import make_subplots
import plotly.graph_objects as go

from hexagon_ml import UCR_Anomaly

from cnndrad import DataReconstructionAD
from cnndrad.utils import Signal

# -----------------------------------------------------------
# TEST
# -----------------------------------------------------------

def run_test(window_split=3, work_dir='.'):
    
    work_dir = Path(work_dir)
    
    dirname = work_dir / 'output'
    if os.path.exists(dirname):
        shutil.rmtree(dirname)
    
    # -------
    # DATASET
    # -------
    
    datasets = UCR_Anomaly()    
    dataset = datasets.available_datasets()[0]
        
    # -- Training and test data --
    
    train_data, test_data = datasets.load_data(dataset)
    
    print(f'Found {len(train_data)} values in train data')
    print(f'Found {len(test_data)} values in test data')
    
    # -- Ground truth for anomaly detection --
    
    gt_start, gt_end = datasets.anomaly_segment(dataset)
    gt_indexes = np.where((gt_start <= test_data[:,0]) & (test_data[:,0] <= gt_end))[0]
    
    # -----------
    # WINDOW SIZE
    # -----------
    
    filename= work_dir / 'output' / 'period.png'
    period, confidence = Signal.compute_period(train_data, plot=False, savefig=True, filename=filename)
    #print(f'period: {period:.2f} - confidence: {confidence:.2f}')
    
    window_size = int(np.ceil(period / window_split) * window_split)
    print(f'window size: {window_size}')
    
    # -----------------
    # ANOMALY DETECTION
    # -----------------
    
    anom_detector = DataReconstructionAD(window_size, window_stride=1, reconstruct=[True]*window_split,
                                         batch_size=128, epochs=50, validation_split=0.2,
                                         work_dir=work_dir, device='/gpu:0')
    
    # -- Learn data representation --
    
    anom_detector.fit(train_data)

    #anom_detector.plot_history(savefig=True)
    #anom_detector.plot_metric_windows_distribution(train_data, savefig=True)
    #anom_detector.compare_model_performance_with_baseline(train_data)
    
    # -- Compute anomaly score --

    #anom_detector.plot_score_windows_distribution(test_data, savefig=True)
    #anom_detector.plot_data_windows_reconstruction(test_data)

    score = - anom_detector.score_samples(test_data)
    
    def plot_score_samples(X, ground_truth, score, window_size):
        fig = make_subplots(rows=2, cols=1)
        fig.append_trace(go.Scatter(x=X[:,0], y=X[:,1], mode='lines', name='Dataset', line_color='rgb(114,125,250)'), row=1, col=1)
        fig.append_trace(go.Scatter(x=ground_truth[:,0], y=ground_truth[:,1], mode='lines', name='Ground Truth', line_color='rgb(44,160,44)'), row=1, col=1)
        fig.append_trace(go.Scatter(x=X[:,0], y=score, mode='lines', name='Anomaly score', line_color='rgb(239,85,59)'), row=2, col=1)
        fig.update_xaxes(fixedrange=False, rangeslider_visible=False, row=1, col=1)
        fig.update_xaxes(fixedrange=False, rangeslider_visible=True, row=2, col=1)
        fig.update_xaxes(matches='x')
        fig.update_layout(title=f'Sliding window size: {window_size}')
        
        filename = work_dir / 'output' / 'score_samples.html'
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        plotly.offline.plot(fig, filename=str(filename), include_plotlyjs='cdn', auto_open=False)
    
    plot_score_samples(test_data, test_data[gt_indexes], score, window_size)
    
    # -----------
    # TEST RESULT
    # -----------
    
    gt_score = max(score[gt_indexes])
    test_result = 'PASSED' if gt_score > 1 else 'FAILED'
    print(f'Test {test_result}')


if __name__ == '__main__':

    work_dir = './results'

    run_test(work_dir=work_dir)
